=== modified file 'bzrlib/dirstate.py'
--- bzrlib/dirstate.py	2007-03-02 03:19:49 +0000
+++ bzrlib/dirstate.py	2007-03-04 05:11:09 +0000
@@ -20,7 +20,7 @@
 lines by NL. The field delimiters are ommitted in the grammar, line delimiters
 are not - this is done for clarity of reading. All string data is in utf8.
 
-MINIKIND = "f" | "d" | "l" | "a" | "r";
+MINIKIND = "f" | "d" | "l" | "a" | "r" | "t";
 NL = "\n";
 NULL = "\0";
 WHOLE_NUMBER = {digit}, digit;
@@ -86,6 +86,8 @@
     sha1 value.
 'l' is a symlink entry: As for directory, but a symlink. The fingerprint is the
     link target.
+'t' is a reference to a nested subtree; the fingerprint is the referenced
+    revision.
 
 
 --- Format 1 had the following different definition: ---
@@ -220,10 +222,32 @@
     A dirstate is a specialised data structure for managing local working
     tree state information. Its not yet well defined whether it is platform
     specific, and if it is how we detect/parameterise that.
+
+    Dirstates use the usual lock_write, lock_read and unlock mechanisms.
+    Unlike most bzr disk formats, DirStates must be locked for reading, using
+    lock_read.  (This is an os file lock internally.)  This is necessary
+    because the file can be rewritten in place.
+
+    DirStates must be explicitly written with save() to commit changes; just
+    unlocking them does not write the changes to disk.
     """
 
-    _kind_to_minikind = {'absent':'a', 'file':'f', 'directory':'d', 'relocated':'r', 'symlink':'l'}
-    _minikind_to_kind = {'a':'absent', 'f':'file', 'd':'directory', 'l':'symlink', 'r':'relocated'}
+    _kind_to_minikind = {
+            'absent': 'a',
+            'file': 'f',
+            'directory': 'd',
+            'relocated': 'r',
+            'symlink': 'l',
+            'tree-reference': 't',
+        }
+    _minikind_to_kind = {
+            'a': 'absent',
+            'f': 'file',
+            'd': 'directory',
+            'l':'symlink',
+            'r': 'relocated',
+            't': 'tree-reference',
+        }
     _to_yesno = {True:'y', False: 'n'} # TODO profile the performance gain
      # of using int conversion rather than a dict here. AND BLAME ANDREW IF
      # it is faster.
@@ -287,17 +311,20 @@
         self._split_path_cache = {}
         self._bisect_page_size = DirState.BISECT_PAGE_SIZE
 
-    def add(self, path, file_id, kind, stat, link_or_sha1):
+    def add(self, path, file_id, kind, stat, fingerprint):
         """Add a path to be tracked.
 
         :param path: The path within the dirstate - '' is the root, 'foo' is the
             path foo within the root, 'foo/bar' is the path bar within foo 
             within the root.
         :param file_id: The file id of the path being added.
-        :param kind: The kind of the path.
+        :param kind: The kind of the path, as a string like 'file', 
+            'directory', etc.
         :param stat: The output of os.lstat for the path.
-        :param link_or_sha1: The sha value of the file, or the target of a
-            symlink. '' for directories.
+        :param fingerprint: The sha value of the file,
+            or the target of a symlink,
+            or the referenced revision id for tree-references,
+            or '' for directories.
         """
         # adding a file:
         # find the block its in. 
@@ -363,7 +390,7 @@
         minikind = DirState._kind_to_minikind[kind]
         if kind == 'file':
             entry_data = entry_key, [
-                (minikind, link_or_sha1, size, False, packed_stat),
+                (minikind, fingerprint, size, False, packed_stat),
                 ] + parent_info
         elif kind == 'directory':
             entry_data = entry_key, [
@@ -371,7 +398,11 @@
                 ] + parent_info
         elif kind == 'symlink':
             entry_data = entry_key, [
-                (minikind, link_or_sha1, size, False, packed_stat),
+                (minikind, fingerprint, size, False, packed_stat),
+                ] + parent_info
+        elif kind == 'tree-reference':
+            entry_data = entry_key, [
+                (minikind, fingerprint, 0, False, packed_stat),
                 ] + parent_info
         else:
             raise errors.BzrError('unknown kind %r' % kind)
@@ -1295,6 +1326,8 @@
         :param tree_index: The index of the tree we wish to locate this path
             in. If the path is present in that tree, the entry containing its
             details is returned, otherwise (None, None) is returned
+            0 is the working tree, higher indexes are successive parent
+            trees.
         :param fileid_utf8: A utf8 file_id to look up.
         :param path_utf8: An utf8 path to be looked up.
         :return: The dirstate entry tuple for path, or (None, None)
@@ -1334,14 +1367,18 @@
                 entry_index, present = self._find_entry_index(key, block)
                 if present:
                     entry = self._dirblocks[block_index][1][entry_index]
-                    if entry[1][tree_index][0] in 'fdl':
+                    if entry[1][tree_index][0] in 'fdlt':
                         # this is the result we are looking for: the  
                         # real home of this file_id in this tree.
                         return entry
                     if entry[1][tree_index][0] == 'a':
                         # there is no home for this entry in this tree
                         return None, None
-                    assert entry[1][tree_index][0] == 'r'
+                    assert entry[1][tree_index][0] == 'r', \
+                        "entry %r has invalid minikind %r for tree %r" \
+                        % (entry,
+                           entry[1][tree_index][0],
+                           tree_index)
                     real_path = entry[1][tree_index][1]
                     return self._get_entry(tree_index, fileid_utf8=fileid_utf8,
                         path_utf8=real_path)
@@ -1407,8 +1444,12 @@
             fingerprint = inv_entry.text_sha1 or ''
             size = inv_entry.text_size or 0
             executable = inv_entry.executable
+        elif kind == 'tree-reference':
+            fingerprint = inv_entry.reference_revision or ''
+            size = 0
+            executable = False
         else:
-            raise Exception
+            raise Exception("can't pack %s" % inv_entry)
         return (minikind, fingerprint, size, executable, tree_data)
 
     def _iter_entries(self):

=== modified file 'bzrlib/tests/test_dirstate.py'
--- bzrlib/tests/test_dirstate.py	2007-03-02 02:11:59 +0000
+++ bzrlib/tests/test_dirstate.py	2007-03-04 05:11:09 +0000
@@ -726,6 +726,35 @@
         finally:
             state.unlock()
 
+    def test_add_tree_reference(self):
+        # make a dirstate and add a tree reference
+        state = dirstate.DirState.initialize('dirstate')
+        expected_entry = (
+            ('', 'subdir', 'subdir-id'),
+            [('t', 'subtree-123123', 0, False,
+              'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx')],
+            )
+        try:
+            state.add('subdir', 'subdir-id', 'tree-reference', None, 'subtree-123123')
+            entry = state._get_entry(0, 'subdir-id', 'subdir')
+            self.assertEqual(entry, expected_entry)
+            state._validate()
+            state.save()
+        finally:
+            state.unlock()
+        # now check we can read it back
+        state.lock_read()
+        state._validate()
+        try:
+            entry2 = state._get_entry(0, 'subdir-id', 'subdir')
+            self.assertEqual(entry, entry2)
+            self.assertEqual(entry, expected_entry)
+            # and lookup by id should work too
+            entry2 = state._get_entry(0, fileid_utf8='subdir-id')
+            self.assertEqual(entry, expected_entry)
+        finally:
+            state.unlock()
+
 
 class TestGetLines(TestCaseWithDirState):
 

